/*
 * formel.cpp
 * wtab version 0.02
 * 
 * Copyright 2019 S. Krause <sebkrause@sebkrause.de>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA.
 * 
 */

#include "formel.hpp"
#include <iostream>

/* Konstruktor - erzeugt Formelobjekte */
Formel::Formel(std::string str)
{
    originalstring = str;
    str = trim(str);
      
    /* Atom? */
    if (istVariable(str))
    {
        vars.insert(str);
        junktor = &junktoren[0];
        istatom = true;
    }
    else
    {
        istatom = false;
    }

    /* Formel. */
    if (!istatom)
    {
        size_t pos = 0, junktornummer = 0, junktorpos = 0, junktorlaenge = 0;
        int prio = -1;
        
        while (pos < str.size())
        {
            /* geklammerten Ausdruck überspringen */
            if (str.at(pos) == '(')
            {
                int klammern = 1;
                pos++;
                while (klammern != 0 && pos < str.size())
                {
                    if (str.at(pos) == '(')
                        klammern++;
                    if (str.at(pos) == ')')
                        klammern--;
                    pos++;
                }
                continue;
            }
            
            else
            {
                /* Steht hier eine Variable? */
                if (istBuchstabe(str.at(pos)))
                {
                    pos++;
                    while(pos < str.size() && istZiffer(str.at(pos)))
                    {
                        pos++;
                    }
                    continue;
                }
                
                /* Es muss sich um einen Junktor handeln */
                bool found = false;
                for (size_t i = 1; i < junktoren.size(); i++)
                {
                    int l = junktoren[i].abbr.size();
                    if (str.substr(pos, l) == junktoren.at(i).abbr)
                    {
                        // Einen Junktor gefunden
                        found = true;
                        junktorlaenge = l;
                        if (prio == -1 || junktoren.at(i).prioritaet < prio)
                        {
                            // ...und der bindet auch schwächer als
                            // eventuelle vorherige!
                            junktornummer = i;
                            junktorpos = pos;
                            prio = junktoren.at(i).prioritaet;
                        }
                    }
                }
                if (found)
                {
                    pos += junktorlaenge;
                    continue;
                }
            }
            /* Nichts passendes; Fehler. */
            throw("Parserfehler! Zeichen " + str.substr(pos, 1) + " nicht erkannt.");
        }
        // Junktor gefunden
        if (junktornummer != 0)
        {
            junktor = &junktoren[junktornummer];
            // Länge des Junktorzeichens
            int l = junktor->abbr.size(); 
            // Teilformeln rechts und links vom Junktor erzeugen neue
            // Formeln. Danach werden sie aus dem String gelöscht, wie
            // auch der Junktor - ist dann noch etwas übrig: Fehler!
            if (junktor->right)
            {
                kindformeln.push_back(new Formel(str.substr(junktorpos + l, str.size() - (junktorpos + l))));
                for (std::string c : kindformeln.back()->vars)
                    vars.insert(c);
                str.erase(junktorpos + l, str.size() - (junktorpos + l));
            }
            str.erase(junktorpos, l);
            if (junktor->left)
            {
                kindformeln.push_back(new Formel(str.substr(0, junktorpos)));
                for (std::string c : kindformeln.back()->vars)
                    vars.insert(c);
                str.erase(0, junktorpos);
            }
            // Irgendetwas wurde nicht verarbeitet...
            if (!str.empty())
            {
                throw("Teilformel '" + str + "' konnte nicht verarbeitet werden!");
            }
            // Da erst rechts, dann links verarbeitet wurde:
            // Tauschen der beiden Seiten
            if (kindformeln.size() == 2)
            {
                Formel * tmp = kindformeln[0];
                kindformeln[0] = kindformeln[1];
                kindformeln[1] = tmp;
            }
        }
        else
        {
            throw("Parserfehler! Kein Operator gefunden.");
        }
    }
}


/* Gibt den Wert der Formel unter einer Belegung zurück */
int Formel::wert(std::map<std::string, int> belegung) const
{
    /* strukturelle Induktion - Anfang */
    if (istatom)
    {
        return belegung[*(vars.begin())];
    }
    
    /* strukturelle Induktion - Schluss */
    switch(junktor->stelligkeit)
    {
    case 0:
        return junktor->funktion(0, 0);
    case 1:
        return junktor->funktion(kindformeln.at(0)->wert(belegung), 0);
    default:
        return junktor->funktion(kindformeln.at(0)->wert(belegung), kindformeln.at(1)->wert(belegung));
    }
}


/* Zeichnet eine Wertetabelle */
void Formel::werteTabelle() const
{
    std::map<std::string, int> belegung;       // p -> 1, q -> 0, x1 -> 0 o.ä.
    std::vector<std::map<std::string, int>> modelle; // Belegungen mit Wert = 1
    int l = originalstring.size() + 2;  // Breite der Formel (für Ausgabe)
    std::map<std::string, int> breite;
    std::set<std::string>::iterator it;
    bool print = (vars.size() > 0);
    
    /* Header schreiben, Belegung vorbereiten */
    print && std::cout << std::endl;
    for (std::string c : vars)
    {
        belegung[c] = 0;                // Belegung initialisieren
        breite[c] = c.size();
        print && std::cout << " " << c << " |";  // Tabellenkopf
        
    }
    print && std::cout << " " << originalstring << std::endl;
    for (it = vars.begin(); it != vars.end(); it++)
        print && std::cout << std::string(breite[*it] + 2, '-')  << "+";            // Trennlinie unter Kopf
    print && std::cout << std::string(l, '-') << std::endl;
    
    // Variablen als Binärzahlstellen zum Hochzählen von 0 bis max-1 */
    unsigned int max = 1;               // 3 Variablen: max = 2^3 = 8
    for (size_t i = 0; i < vars.size(); i++)
        max *= 2;
        
    /* Wertetabelle: */
    
    for (unsigned int val = 0; val < max; val++)
    {
        int i = vars.size() - 1; // vorderste Variable zuletzt
        for (it = vars.begin(); it != vars.end(); it++)
        {
            belegung[*it] = (val >> i) & 1;
            i--;
            print && std::cout 
                << std::string(breite[*it] / 2 + 1, ' ')
                << belegung[*it] 
                << std::string(breite[*it] / 2 + (breite[*it] % 2), ' ')
                << '|'
            ;
        }
        int w = wert(belegung);
        if (w == 1)
            modelle.push_back(belegung);
        print && std::cout << std::string(l/2, ' ') << w << std::endl;
    }
    std::cout << std::endl;
    
    /* Kontradiktion / Tautologie */
    if (modelle.size() == 0)
        std::cout << "Kontradiktion! Keine Belegung erfüllt die Formel." << std::endl;
    else if (modelle.size() == max)
        std::cout << "Tautologie! Jede Belegung erfüllt die Formel." << std::endl;
    
    /* Variablenmenge */
    std::cout << "Variablenmenge: vars = ";
    if (vars.size() == 0)
        std::cout << "leer" << std::endl;
    else
    {
        std::cout << "{";
        size_t i = 0;
        for (it = vars.begin(); it != vars.end(); it++)
        {
            std::cout << *it;
            if (i < vars.size() - 1)
                std::cout << ", ";
            i++;
        }
        std::cout << "}" << std::endl;
    }
    
    /* Modellmenge */
    if (print)
    {
        std::cout << "Modellmenge: Mod = {";
        for (size_t i = 0; i < modelle.size(); i++)
        {
            belegung = modelle.at(i);
            std::cout << "W_";
            for (std::map<std::string, int>::iterator mit = belegung.begin(); mit != belegung.end(); mit++)
                std::cout << mit->second;
            if (i < modelle.size() - 1)
                std::cout << ", ";
        }
        std::cout << "}" << std::endl;
    }
}


/* Entfernt alle Leerzeichen und äußere Klammerpaare */
std::string Formel::trim(std::string str)
{
    std::string res = "";
    /* Leerzeichen löschen */
    for (size_t i = 0; i < str.size(); i++)
        if (str.at(i) != ' ')
            res += str.at(i);
    /* Äußerste Klammerpaare löschen */
    while (res.size() > 0 && res.at(0) == '(' && res.at(res.size() - 1) == ')')
    {
        int klammern = 1;
        size_t pos = 1;
        while (pos < res.size())
        {
            if (res.at(pos) == '(')
                klammern++;
            else if (res.at(pos) == ')')
                klammern--;
            pos++;
            if (klammern == 0)
                break;
        }
        if (pos == res.size())
            res = res.substr(1, res.size() - 2);
        else
            break;
    }
    return res;
}

/* Ist String s eine mögliche Variable? */
bool Formel::istVariable(const std::string& s)
{
    /* Ist s leer? */
    if (s.size() == 0)
        return false;
    /* Ist erstes Zeichen ein Buchstabe? */
    if (istBuchstabe(s.at(0)))
    {
        /* Sind (falls vorhanden) alle anderen Zeichen Ziffern? */
        for (size_t p = 1; p < s.size(); p++)
        {
            if (!istZiffer(s.at(p)))
                return false;
        }
        return true;
    }
    return false;
}

/* Ist Zeichen c ein Buchstabe? */
bool Formel::istBuchstabe(char c)
{
    if (c >= 'a' && c <= 'z')
        return true;
    if (c >= 'A' && c <= 'Z')
        return true;
    return false;
}

/* Ist Zeichen c eine Ziffer? */
bool Formel::istZiffer(char c)
{
    if (c >= '0' && c <= '9')
        return true;
    return false;
}
