/*
 * junktor.hpp
 * wtab version 0.02
 * 
 * Copyright 2019 S. Krause <sebkrause@sebkrause.de>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA.
 * 
 */
 
#pragma once

#include <string>
#include <vector>

/* Die Datenstruktur für Junktoren */
struct junkt
{
    // Name des Junktors
    std::string name;
    // Abkürzung: Das, was der User eingeben muss
    std::string abbr;
    // Stelligkeit des Junktors:
    //      --> 0 im Falle v. true und false
    //      --> 1 im Falle von not
    //      --> 2 sonst
    unsigned int stelligkeit;
    // Funktionszeiger auf die Funktion, die den Wert errechnet
    int (*funktion)(int, int);
    // Wird links vom Junktor eine Formel erwartet?
    bool left;
    // Wird rechts vom Junktor eine Formel erwartet?
    bool right;
    // Die Bindungspriorität. ! bindet stärker als &&, hat also
    // eine höhere Priorität.
    int prioritaet;
};

/* Die Funktionen für die Junktoren
 * Maximale Stelligkeit = 2, also 2 Parameter */
static int jAtom   (int a,     int b = 0)  { return a; }
static int jTrue   (int a = 0, int b = 0)  { return 1; }
static int jFalse  (int a = 0, int b = 0)  { return 0; }
static int jNot    (int a,     int b = 0)  { return 1 - a; }
static int jAnd    (int a,     int b)      { return std::min(a, b); }
static int jXor    (int a,     int b)      { return (a != b) ? 1 : 0; }
static int jOr     (int a,     int b)      { return std::max(a, b); }
static int jImpl   (int a,     int b)      { return (a <= b) ? 1 : 0; }
static int jEquiv  (int a,     int b)      { return (a == b) ? 1 : 0; }
static int jNand   (int a,     int b)      { return 1 - std::min(a, b); }
static int jNor    (int a,     int b)      { return 1 - std::max(a, b); }

/* Das Array der Junktoren incl. Zeiger auf die Funktionen */
static std::vector<junkt> junktoren =
{
    // Name     Syntax  Stellig-    Funkt.- links   rechts  Bindungs-
    //                  keit        zeiger                  priorität
    { "ATOM",   "",     1,          jAtom,  true,   false,  10000 },
    { "TRUE",   "1",    0,          jTrue,  false,  false,  1000 },
    { "FALSE",  "0",    0,          jFalse, false,  false,  1000 },
    { "NOT",    "!",    1,          jNot,   false,  true,   900 },
    { "AND",    "&&",   2,          jAnd,   true,   true,   800 },
    { "XOR",    "+",    2,          jXor,   true,   true,   750 },
    { "OR",     "||",   2,          jOr,    true,   true,   700 },
    { "IMPL",   "=>",   2,          jImpl,  true,   true,   600 },
    { "EQUIV",  "<=>",  2,          jEquiv, true,   true,   500 },
    { "NAND",   "&!&",  2,          jNand,  true,   true,   775 },
    { "NOR",    "|!|",  2,          jNor,   true,   true,   650 }
};
