/*
 * main.cpp
 * wtab version 0.02
 * 
 * Copyright 2019 S. Krause <sebkrause@sebkrause.de>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA.
 * 
 */

#include "formel.hpp"

#include <iostream>
#include <string>
#include <exception>

/* liest eine Zeile Usereingabe aus std::cin (inkl. Leerzeichen) */
std::string inputString()
{
    std::string result = "";
    char c = '\0';
    while (c != '\n' && c != '\r')
    {
        if (c != '\0')
        {
            result = result + c;
        }
        c = (char) std::cin.get();
    }
    return result;
}

/* Hilfetext zeigen */
void help()
{
    std::cout <<
        "Wertetabellen aussagenlogischer Formeln.\n"
        "Syntax:\n"
        "   - Variablen: Möglich sind alle Buchstaben von a-z und A-Z,\n"
        "     gefolgt von beliebig vielen Ziffern. Gültig also\n"
        "     a, P, x1, y234\n"
        "   - Operatoren: Zur Verfügung stehen\n"
        "           1     TRUE              0     FALSE\n"
        "           !     NOT               &&    UND\n"
        "           &!&   NAND              +     XOR\n"
        "           ||    OR                |!|   NOR\n"
        "           =>    IMPLIKATION       <=>   ÄQUIVALENZ\n"
        "     in absteigender Bindungsreihenfolge. D.h. die Formel\n"
        "     \"a &!& b && !c\" wird interpretiert als\n"
        "     \"a &!& ( b && ( !c ) )\"\n"
        "   - Klammern: Zur Verfügung stehen nur runde Klammern ( und ).\n"
        "   - Leerzeichen werden ignoriert (können also genutzt werden).\n"
        "Beispiel:  Eingabe 'a => (a && b)'. Ausgabe:\n"
        " a | b | a => (a && b)\n"
        "---+---+--------------\n"
        " 0 | 0 |       1\n"
        " 0 | 1 |       1\n"
        " 1 | 0 |       0\n"
        " 1 | 1 |       1\n"
        "\n"
        "Variablenmenge: vars = {a, b}\n"
        "Modellmenge: Mod = {W_00, W_01, W_11}\n"
        "\n"
        "(Die Variablenmenge wird angegeben, damit die Modellindizes\n"
        "hier eindeutig bestimmt sind.)\n"
        "S. Krause 2019 <sebkrause@sebkrause.de>\n"
    ;
}

/* Hauptprogramm */
int main()
{
    Formel *f;
    
    while (true)
    {
        /* Ausgaben */
        std::cout << std::endl
                  << ":q - Ende, :h - Hilfe" << std::endl
                  << "Formel eingeben:  "
        ;
        
        /* Usereingabe entgegennehmen */
        std::string s = inputString();
        
        /* Auswerten: */
        
        /* --> Beenden */
        if (s == ":q")
        {
            std::cout << "Programm beendet." << std::endl;
            return 0;
        }
        
        /* --> Hilfe */
        if (s == ":h")
        {
            help();
            continue;
        }
        
        /* --> Formel einlesen */
        try
        {
            f = new Formel(s);  // Formel erzeugen
            f->werteTabelle();  // Wertetabelle anzeigen
            free(f);            // löschen
        }
        catch ( std::string e )
        {
            std::cout << "Fehler: " << e << std::endl;
        }
    }
    return 0;
}
